
package org.omegahat.DataStructures.Data;

import java.util.Vector;

import org.omegahat.Environment.Tools.DataScanner.RecordStreamListener;
import org.omegahat.Environment.Tools.DataScanner.ObjectReader;


import org.omegahat.Environment.Parser.Parse.List;

import org.omegahat.Environment.Language.Function;

import org.omegahat.Environment.Interpreter.Evaluator;




/**
  Early version of a class that stores arbitrary records in a list,
that is row-wise, 
in contrast to a DataFrame that stores variables
(that is columnwise).
This supports notification from a stream source which 
provides each record and an object of this class can 
process the record before storing it or simply discard it.
*/

public class DataList extends java.util.Vector 
            implements RecordStreamListener,
                       org.omegahat.Environment.DataStructures.Subsettable,
                       org.omegahat.Environment.DataStructures.AssignableSubset,
                       java.io.Serializable
{
   static final public int ROWS = 0;
   static final public int COLUMNS = 1;

   
   protected Vector names[] = new java.util.Vector[2];

    protected Object[] data;
    protected int currentNum = 0;

  
  public DataList() {
    
  }

  public DataList(Object[] els) {
    data(els);
  }

  public DataList(int size) {
    super(size);
    data = new Object[size];
  }


  public DataList(ObjectReader reader) {
    reader.addListener((RecordStreamListener)this);
  }

  public DataList(ObjectReader reader, int size) {
    this(size);
    reader.addListener((RecordStreamListener)this);
  }

  public DataList(java.io.File f) throws java.io.FileNotFoundException {
   this(new org.omegahat.Environment.Tools.DataScanner.EventDataScanner(new java.io.FileReader(f)));
  }


  public int data(Object els[]) {
    for(int i = 0 ; i < els.length; i++) {
      addElement(els[i]);
    }
   return(size());
  }


  public boolean newRecord(Object obj, ObjectReader source) {
    // addRecord(obj);
    obj = processElement(obj);
    if(obj != null) {
          super.addElement(obj);
      //    data[currentNum++] = obj;
    }

  return(true);
  }

  
  public void addRecord(Object obj) {
    obj = processElement(obj);
    if(obj != null) {
      super.addElement(obj);
    }
  }

  public Object processElement(Object obj) {
   return(obj);
  }



  public Vector rowNames() {
   return(names(ROWS));
  }

  public Vector rowNames(Vector v) {
    // need to do some checking on length here!
   return(names(ROWS, v));
  }


  public Vector rowNames(String[] vals) {
   return(names(ROWS, toVector(vals)));
  }

  public Vector columnNames() {
   return(names(COLUMNS));
  }

  public Vector columnNames(Vector v) {
    // need to do some checking on length here!
   return(names(COLUMNS,v));
  }


  public Vector columnNames(String[] vals) {
   return(names(ROWS, vals));
  }

  public Vector names(int which) {
   return(names[which]);
  }

  public Vector names(int which, Vector v) {
   names[which] = v;
   return(names(which));
  }

  public Vector names(int which, String[] vals) {
   return(names(which, toVector(vals)));
  }

  public Vector toVector(String[] vals) {
    if(vals == null) 
      return(null);

    // need to do some checking on length here!
    Vector v = new Vector(vals.length);
    for(int i = 0 ; i < vals.length ;i++) {
      v.addElement(vals[i]);
    }

  return(v);
  }



  public DataList apply(Function f) {
   try {
     return(apply(f, org.omegahat.Environment.System.Globals.evaluator()));
   } catch(ClassNotFoundException ex) {
     throw new RuntimeException("Can't obtain an evaluator");
   }
  }

  public DataList apply(Function f, Vector data) {
   try {
     return(apply(f, org.omegahat.Environment.System.Globals.evaluator(), data));
   } catch(ClassNotFoundException ex) {
     throw new RuntimeException("Can't obtain an evaluator");
   }
  }

  public DataList apply(Function f, Evaluator evaluator) {
    return(apply(f, evaluator, this));
  }

  public DataList apply(Function f, Evaluator evaluator, Vector data) {
    DataList v = new DataList(data.size());
    Object record, val;
    List arg;
    for(int i = 0; i < data.size(); i++) {   
      record = data.elementAt(i);
      arg = createArgumentList(record, f);
      val = null;
      try {
        val = f.eval(arg, evaluator);
      } catch(Throwable e) {
        evaluator.warning("Apply failed on element " + i);
        val = org.omegahat.Environment.Databases.Database.NULL_ENTRY;
      }
        v.addElement(val);
    }

   return(v);
  }

  protected List createArgumentList(Object obj, org.omegahat.Environment.Language.Function f) {
   String argName = (String) f.argumentNames().elementAt(0);

   List list = new List((Object)new org.omegahat.Environment.Parser.Parse.AssignExpression(
                                              new org.omegahat.Environment.Parser.Parse.Name(argName),obj)); 

   return(list);
  }


  public DataList apply(org.omegahat.Environment.Parser.Parse.MethodCall method, Evaluator evaluator) {
    return(apply(method, evaluator,  new List(0)));
  }

  public DataList apply(org.omegahat.Environment.Parser.Parse.MethodCall method, Evaluator evaluator, List args) {
    DataList v = new DataList(size());
    Object record, val;
    for(int i = 0; i < size(); i++) {   
      record = elementAt(i);
      val = null;
      try {
        val = method.findEvalMethod(args, record, evaluator, false);
      } catch(Throwable e) {
        evaluator.warning("Apply failed on element " + i);
  e.printStackTrace();
        val = org.omegahat.Environment.Databases.Database.NULL_ENTRY;
      }
        v.addElement(val);
    }

   return(v);
  }


  public Object subset(Object index, boolean keepStructure) throws Throwable {

    Object obj = null;
    if(index instanceof List) {
      FilterDataList subset = new FilterDataList(this);
      int n = ((List)index).size();
      for(int i = 0; i < n; i++) {
        obj = ((List)index).elementAt(i);
        subset.addElement(subset(obj, false));
      }
      if(((List)index).size() == 1) 
       obj = subset.elementAt(0);
      else
       obj = subset;
    } else if(index instanceof org.omegahat.Environment.Language.Function) {
       obj = new FilterDataList(this, (org.omegahat.Environment.Language.Function)index);
    }
  /*
   else if(index instanceof org.omegahat.Environment.Parser.Parse.ExpressionInt) {
       subset = new FilterDataList(this, (org.omegahat.Environment.Parser.Parse.ExpressionInt)index);
    }
  */
    else if(index instanceof org.omegahat.Environment.Language.Evaluable) {
   
      try {
        Object tmp = ((org.omegahat.Environment.Language.Evaluable)index).eval(org.omegahat.Environment.System.Globals.evaluator());
       obj = subset(tmp, keepStructure);
      } catch(ClassNotFoundException ex) {
          throw new RuntimeException("Cannot obtain an evaluator for subset()");
      }
    } else if(index instanceof Number) {
      int idx = ((Number)index).intValue();
      obj = elementAt(idx);
    }

  return(obj);
  }

  public Object assignSubset(Object index, Object value) {

  return(value);
  }


}

/*
Copyright (c) 1998, 1999 The Omega Project for Statistical Computing.
     All rights reserved.

     Redistribution and use in source and binary forms, with or without
     modification, are permitted provided that the following conditions are
     met:

       Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

       Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

       Neither name of the Omega Project for Statistical Computing nor the 
       names of its contributors may be used to endorse or promote products 
       derived from this software without specific prior written permission.

     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY
     EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
     SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
     SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
     PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
     INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
     STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
     THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

